;;; b.el --- major mode for editing B source code in Emacs

;; Copyright (C) 2002 cse@UNSW

;; Author: Thai Son Hoang (htson@cse.unsw.edu.au)
;; Keywords: languages

;; This file is NOT part of XEmacs.

;; XEmacs is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; XEmacs is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with XEmacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; USAGE
;; =====

;; Add the following to your Emacs customization file (usually .emacs) (You can
;; also associate other 'B' files (.ref, .imp, .itf) with this mode).
;; (setq auto-mode-alist
;;      (append '(("\\.mch$"       . b-mode))
;;	      auto-mode-alist))


;; Emacs should enter  mode when you find a B specification source file (.mch)
;; When you have entered B mode, you may get more info by pressing
;; C-h m. You may also get online help describing various functions by:
;; C-h f <Name of function you want described>.

;; If you want to customize B mode to fit you better, you may add
;; these lines (the values of the variables presented here are the defaults):
;;
;; ;; User customization for B mode
;; (setq b-indent-level       2
;;       b-tab-always-indent  t)

;; FIXES
;; Jul 11, 02 - 1.0: Original version
;; Aug 27, 02 - 1.1: Fixed the indentation for comments in B.
;; Aug 28, 02 - 1.2: Fixed insert block commands.
;; Aug 29, 02 - 1.3: Fixed comments indentation problem.
;; Aug 30, 02 - 1.4: Added few insert block commands (Operation comments,
;;                   header, etc.). b-save-buffer which will notify the
;;                   modified date and before saving the buffer
;; Sep 10, 02 - 2.0: Added few highlight syntaxes for Refinement and
;;                   Implementation.
;; Nov 20, 02 - 2.1: Added few highlight syntaxex for BaseGenerator, Interface,
;;                   and ProofMethod files.
;; Nov 21, 02 - 2.2: Modify some insert block commands.
;;                   Modify insert operation block to take inputs
;; Nov 26, 02 - 2.3: Add few latex macro function for editing printable comment
;; Dec 11, 02 - 2.4: Fix some bugs with latex macros.
;; Mar 06, 03 - 2.5: Fix some bugs with latex macros.

;; KNOWN BUGS / BUGREPORTS
;; =======================
;;
;; The indentation is not flexible, in other word, if you are using different
;; style, it may not work. I'm not sure is that the right way to put it, but I 
;; feel there should be a standard indentation that you should follow.
;;
;; TO DO NEXT
;; ==========
;;
;; The current version 2.5 has the hightlight syntax for the specification,
;; refinement and implementation. At last it can handle that.
;;
;; I also have the idea of outline mode for B but haven't got time to look at
;; it.
;;
;; THANKS
;; ======
;;
;; I have to thank my supervisor Ken Robinson (kenr@cse.unsw.edu.au) for his
;; helping at this very first stage, including his suggestion and consultation
;; in terms of understanding Emacs-Lisp.
;;

;; added by mal:
(require 'font-lock)



;;; Code:
(setq texmacro "emph")
(setq lvar "bb")

(defconst b-mode-version "2.4"
  "Version of `b.el'.")

(defgroup b nil
  "Major mode for editing B source in Emacs"
  :group 'languages)

(defvar b-mode-map ()
  "Keymap used in B mode.")

;;
;; Key maps for b mode
;;
(if b-mode-map
    ()
  (setq b-mode-map (make-sparse-keymap))
  (define-key b-mode-map "\t"       'electric-b-tab)
  (define-key b-mode-map "\C-c\C-b" 'b-insert-begin-block)
  (define-key b-mode-map "\C-c\C-p" 'b-insert-pre-block)
  (define-key b-mode-map "\C-c\C-i" 'b-insert-if-block)
  (define-key b-mode-map "\C-c\C-a" 'b-insert-any-block)
  (define-key b-mode-map "\C-c\C-l" 'b-insert-let-block)
  (define-key b-mode-map "\C-c\C-s" 'b-insert-select-block)
  (define-key b-mode-map "\C-c\C-c" 'b-insert-choice-block)
  (define-key b-mode-map "\C-c\C-e" 'b-insert-case-block)
  (define-key b-mode-map "\C-c\C-m" 'b-insert-unprint-comment-block)
  (define-key b-mode-map "\C-c\C-n" 'b-insert-print-comment-block)
  (define-key b-mode-map "\C-c\C-o" 'b-insert-operation-comment-block)
  (define-key b-mode-map "\C-c\C-h" 'b-insert-header-comment-block)
  (define-key b-mode-map "\C-c\C-v" 'b-insert-var-block)
  (define-key b-mode-map "\C-c\C-w" 'b-insert-while-block)
  (define-key b-mode-map "\M-\C-e"  'b-insert-elsif-substatement)
  (define-key b-mode-map "\M-\C-w"  'b-insert-when-substatement)  
  (define-key b-mode-map "\M-\C-o"  'b-insert-or-substatement)  
  (define-key b-mode-map "\M-\C-t"  'b-insert-or-then-substatement)
  (define-key b-mode-map "\C-n\C-o" 'b-insert-op-block)
  (define-key b-mode-map "\C-n\C-t" 'b-insert-set-block)
  (define-key b-mode-map "\C-n\C-c" 'b-insert-constant-block)
  (define-key b-mode-map "\C-n\C-p" 'b-insert-property-block)
  (define-key b-mode-map "\C-n\C-v" 'b-insert-variable-block)
  (define-key b-mode-map "\C-n\C-i" 'b-insert-invariant-block)
  (define-key b-mode-map "\C-n\C-n" 'b-insert-initialisation-block)
  (define-key b-mode-map "\C-n\C-d" 'b-insert-def-block)
  (define-key b-mode-map "\C-n\C-s" 'b-insert-see-block)
  (define-key b-mode-map "\C-n\C-u" 'b-insert-use-block)
  (define-key b-mode-map "\C-n\C-l" 'b-insert-include-block)
  (define-key b-mode-map "\C-n\C-e" 'b-insert-extend-block)
  (define-key b-mode-map "\C-n\C-m" 'b-insert-machine-block)
  (define-key b-mode-map "\C-n\C-r" 'b-insert-import-block)
  (define-key b-mode-map "\C-j"     'electric-b-new-indent-line)
  (define-key b-mode-map "\C-x\C-s" 'b-save-buffer)
  (define-key b-mode-map "_"        'b-latex-underscore)
  (define-key b-mode-map "\""       'b-latex-doublequote)
  (define-key b-mode-map "\\"       'b-latex-backslash)
  (define-key b-mode-map " "        'b-latex-space)
  )

;;;;
;;;; Regular expressions used to calculate indent, etc.
;;;;
(defconst b-construct-re      "^[ \t]*\\<\\(MACHINE\\|REFINEMENT\\|IMPLEMENTATION\\|DOCUMENT\\|INTERFACE\\|SYSTEM\\|PROOFMETHOD\\|SUPPORTS\\|CONTENTS\\|CONSTRAINTS\\|REFINES\\|SEES\\|USES\\|INCLUDES\\|EXTENDS\\|PROMOTES\\|IMPORTS\\|SETS\\|CONSTANTS\\|PROPERTIES\\|VARIABLES\\|ASSERTIONS\\|EXPECTATIONS\\|INITIALISATION\\|DEFINITIONS\\|OPERATIONS\\|CONTENTS\\|IS\\|INCLUSIONS\\)\\($\\|[ \t]+$\\|[ \t]+[a-z0-9]+\\)")
(defconst b-invariant-re      "^[ \t]*\\<\\(INVARIANT\\)\\($\\|[ \t]+$\\|[ \t]+[a-z0-9]+\\)")
(defconst b-begin-block-re      "^[ \t]*\\<\\(USETAC\\|BASE\\|GLOBAL\\|PRE\\|CHOICE\\|BEGIN\\|IF\\|WHERE\\)\\>")
(defconst b-reason-re     "^[ \t]*\\<\\(REASONING_DEPTH\\)\\>")
(defconst b-end-block-re      "^[ \t]*\\(}\\|)[^\n]*\\|END\\|END[ \t]*;\\|END[ \t]*||\\)[ \t]*$")
;(defconst b-proof-rule-re "^[a-z0-9_,? \t]*\\(.(\\)[ \t]*$")
(defconst b-op-param-out-re      "^[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*\\(<--\\)[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*(\\([a-z0-9_,? \t]*\\))[ \t]*=[ \t]*$")
(defconst b-op-out-re      "^[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*\\(<--\\)[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*=[ \t]*$")
(defconst b-op-param-re      "^[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*(\\([a-z0-9_,? \t]*\\))[ \t]*=[ \t]*$")
(defconst b-op-re      "^[ \t]*\\([a-z0-9_,\? \t]*\\)[ \t]*=[ \t]*$")
(defconst b-then-re    "^[ \t]*\\(MANDATORY\\|OPTIONAL\\|THEN\\|=>\\|IS\\|DO\\|VARIANT\\|BOUND\\|ELSE\\|ELSIF\\|IN\\|OR[^@]*\\|WHEN[^@]*\\)[ \t]*$")
(defconst b-where-re   "^[ \t]*\\(WHERE\\|ON\\|ON[a-z0-9_,? \t]*\\)[ \t]*$")
(defconst b-pchoice-re "^[ \t]*\\(PCHOICE\\|ACHOICE\\)\\>")
(defconst b-begin-comment-re "^[ \t]*/\\(\\*\\|\\*\"\\)[ \t]*$")
(defconst b-begin-print-comment-re "^[ \t]*/\\*\"[ \t]*$")
(defconst b-end-comment-re "^[ \t]*\\(\\*\\|\"\\*\\)/[ \t]*$")
(defconst b-end-print-comment-re "^[ \t]*\"\\*/[ \t]*$")
(defconst b-empty-line-re "^[ \t]*$")

;;;;
;;;; Fontifying
;;;;
(defvar b-font-lock-keywords (purecopy
  (list
   '("^[ \t]*\\(MACHINE\\|REFINEMENT\\|IMPLEMENTATION\\|DOCUMENT\\|INTERFACE\\|SYSTEM\\|PROOFMETHOD\\)[ \t]*\\(\\[a-z]\\)?"
     1 font-lock-preprocessor-face)
   '("^[ \t]*\\(MACHINE\\|REFINEMENT\\|IMPLEMENTATION\\|DOCUMENT\\|INTERFACE\\|SYSTEM\\|PROOFMETHOD\\)\\>[ \t]*\\([a-z][a-z0-9_]*\\)"
     2 font-lock-function-name-face)
   '("^[ \t]*\\(MACHINE\\|INTERFACE\\)\\>[ \t]*\\([a-z][a-z0-9_]*\\)(\\([a-z0-9_, \t]*\\))"
     3 font-lock-variable-name-face)
   '("^[ \t]*\\(CONSTRAINTS\\|REFINES\\|SEES\\|USES\\|INCLUDES\\|PROMOTES\\|EXTENDS\\|IMPORTS\\|SETS\\|CONSTANTS\\|PROPERTIES\\|VARIABLES\\|ASSERTIONS\\|INVARIANT\\|EXPECTATIONS\\|INITIALISATION\\|DEFINITIONS\\|OPERATIONS\\|CONTENTS\\|SUPPORTS\\|IS\\|INCLUSIONS\\)\\>[ \t]*"
     1 font-lock-preprocessor-face)
   (cons (concat "\\<\\(nat\\|nat1\\|bool\\|string\\|integer\\|scalar\\)\\>")
	 'font-lock-type-face)
   (cons (concat "\\<\\(pow\\|pow1\\|fin\\|set\\|seq\\|iseq\\)\\>")
	 'font-lock-warning-face)
   (concat "[ \t]+\\<\\("
	   "pre\\|begin\\|case\\|any\\|where\\|then\\|end\\|else\\|elsif\\|if\\|let\\|be\\|in\\|select\\|choice\\|pchoice\\|achoice\\|or\\|when\\|var\\|of\\|either\\|while\\|do\\|variant\\|bound\\|global\\|base\\|mandatory\\|optional\\|is\\|usetac\\|reasoning_depth\\|on\\|theory\\|"
	   "\\)\\>[ \t]*")
   '("^\\(end\\)" 1 font-lock-preprocessor-face)
   '("^[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*\\(<--\\)[ \t]*" 1 font-lock-variable-name-face)
   '("^[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*\\(<--\\)[ \t]*" 2 font-lock-reference-face)
   '("^[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*\\(<--\\)[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*(\\([a-z0-9_,? \t]*\\))[ \t]*=[ \t]*$" 3 font-lock-function-name-face)
   
   '("^[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*\\(<--\\)[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*(\\([a-z0-9_,? \t]*\\))[ \t]*=[ \t]*$" 4 font-lock-variable-name-face)
   
   '("^[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*(\\([a-z0-9_,? \t]*\\))[ \t]*=[ \t]*$" 1 font-lock-function-name-face)
   
   '("^[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*(\\([a-z0-9_,? \t]*\\))[ \t]*=[ \t]*$" 2 font-lock-variable-name-face)
   
   '("^[ \t]*\\([a-z0-9_,? \t]*\\)[ \t]*=[ \t]*$" 1 font-lock-function-name-face)
   
   '("/\\(\\*\\|\\*\"\\)\\([.*?\n]*\\)\\(\\*\\|\"\\*\\)/" 2 font-lock-string-face)
   ))
  "Additional expressions to highlight in B mode.")

(put 'b-mode 'font-lock-defaults '(b-font-lock-keywords nil t))


;;
;; B mode customize
;;
(defcustom b-indent-level 2
  "*Indentation of B statements with respect to containing block."
  :type 'integer
  :group 'b)

(defcustom b-tab-always-indent t
  "*Non-nil means TAB in B mode should always reindent the current line,
regardless of where in the line point is when the TAB command is used."
  :type 'boolean
  :group 'b)

(defcustom author "Thai Son Hoang"
  "*Name of the author. This will be inserted within the comments of each B
files."
  :type 'string
  :group 'b)

(defcustom email "htson@cse.unsw.edu.au"
  "*Email of the author. This will be inserted within the comments of each B
files."
  :type 'string
  :group 'b)

;(setq author "Thai Son Hoang")
;(setq email "htson@cse.unsw.edu.au")

;;;;###autoload
(defun b-mode ()
  "Major mode for editing b code. \\<pascal-mode-map>
TAB indents for Pascal code.  Delete converts tabs to spaces as it moves back.

Other useful functions are:

\\[\"\\C-c\\C-b\" b-insert-begin-block]\t- insert BEGIN...END block
\\[\"\\C-c\\C-p\" b-insert-pre-block]\t- insert PRE...THEN ...END block
\\[\"\\C-c\\C-i\" b-insert-if-block]\t- insert IF...THEN...ELSE...END block
\\[\"\\C-c\\C-a\" b-insert-any-block]\t- insert ANY...WHERE...THEN...END block
\\[\"\\C-c\\C-l\" b-insert-let-block]\t- insert LET...BE...IN...END block
\\[\"\\C-c\\C-s\" b-insert-select-block]\t- insert SELECT...THEN...WHEN...THEN...ELSE...END block
\\[\"\\C-c\\C-c\" b-insert-choice-block]\t- insert CHOICE...OR...END block
\\[\"\\C-c\\C-e\" b-insert-case-block]\t- insert CASE block
\\[\"\\C-c\\C-m\" b-insert-unprint-comment-block]\t- insert unprintable comment block
\\[\"\\C-c\\C-n\" b-insert-print-comment-block]\t- insert printable comment block
\\[\"\\C-c\\C-v\"  b-insert-var-block]\t- insert VAR...IN...END statement
\\[\"\\C-c\\C-w\"  b-insert-while-block]\t- insert WHILE...DO...VARIANT...INVARIANT...END statement
\\[\"\\M-\\C-e\"  b-insert-elsif-substatement]\t- insert ELSIF substatement
\\[\"\\M-\\C-w\"  b-insert-when-substatement]\t- insert WHEN substatement
\\[\"\\M-\\C-o\"  b-insert-or-substatement]\t- insert OR substatement  
\\[\"\\M-\\C-t\"  b-insert-or-then-substatement]\t- insert OR...THEN substatement
\\[\"\\C-n\\C-o\" b-insert-op-block]\t- insert new operation template
\\[\"\\C-n\\C-t\" b-insert-set-block]\t- insert new SETS template
\\[\"\\C-n\\C-c\" b-insert-constant-block]\t- insert new CONSTANTS template
\\[\"\\C-n\\C-p\" b-insert-property-block]\t- insert new PROPERTIES template
\\[\"\\C-n\\C-v\" b-insert-variable-block]\t- insert new VARIABLES template
\\[\"\\C-n\\C-i\" b-insert-invariant-block]\t- insert new INVARIANT template
\\[\"\\C-n\\C-n\" b-insert-initialisation-block]\t- insert new INITIALISATION template
\\[\"\\C-n\\C-d\" b-insert-def-block]\t- insert new DEFINITIONS template
\\[\"\\C-n\\C-s\" b-insert-see-block]\t- insert new SEES template
\\[\"\\C-n\\C-u\" b-insert-use-block]\t- insert new USES template
\\[\"\\C-n\\C-l\" b-insert-include-block]\t- insert new INCLUDES template
\\[\"\\C-n\\C-e\" b-insert-extend-block]\t- insert new EXTENDS template
\\[\"\\C-n\\C-m\" b-insert-machine-block]\t- insert new MACHINE template
\\[\"\\C-n\\C-p\" b-insert-import-block]\t- insert new IMPORTS template
\\[\"\\C-j\"      electric-b-new-indent-line]\t- new indented line

Variables controlling indentation/edit style:
 b-indent-level      (default 4)
    Indentation of B statements with respect to containing block.
 b-tab-always-indent (default t)
    Non-nil means TAB in Pascal mode should always reindent the current line,
    regardless of where in the line point is when the TAB command is used.

Turning on b mode calls the value of the variable b-mode-hook with
no args, if that value is non-nil."

  (interactive)
  (use-local-map b-mode-map)
  (setq major-mode 'b-mode)
  (setq mode-name "B")
  (setq comment-start "/*")
  (setq comment-end "*/")
  ;; Font lock support
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(b-font-lock-keywords nil t))
  (run-hooks 'b-mode-hook)
  )



;;;;
;;;;  Electric functions
;;;;
(defun electric-b-terminate-line ()
  "Terminate line and indent next line."
  (interactive)
  (delete-horizontal-space) ; Removes trailing whitespaces
  (newline)
  (b-indent-line)
  (end-of-line)
  )

(defun electric-b-new-indent-line ()
  "Terminate the line and indent next line."
  (interactive)
  (newline)
  (b-indent-line)
  (end-of-line)
  )

(defun electric-b-tab ()
  "Function called when TAB is pressed in B mode."
  (interactive)
  (if b-tab-always-indent
      (save-excursion
	(beginning-of-line)
	(b-indent-line)
	)
    (if (save-excursion
	  (skip-chars-backward " \t")
	  (bolp))
	(b-indent-line)
      (insert "\t")
      )
    )
  (beginning-of-line)
  (skip-chars-forward " \t")
  )


;;;;
;;;; Interactive functions
;;;;
(defun b-insert-begin-block ()
  "Insert B BEGIN ... END; block in the code with right indentation."
  (interactive)
  (insert "BEGIN")
  (b-indent-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "END;")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-pre-block ()
  "Insert B PRE ... THEN ... END; block in the code with right indentation."
  (interactive)
  (insert "PRE")
  (b-indent-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "preds?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "THEN")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "END;")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-if-block ()
  "Insert B IF ... THEN ... ELSE ... END; block in the code with right indentation."
  (interactive)
  (insert "IF")
  (b-indent-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "preds?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "THEN")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "ELSE")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "END")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-any-block ()
  "Insert B ANY ... WHERE ... THEN ... END; block in the code with right indentation."
  (interactive)
  (let (
	(newlvar (read-string (concat "Local variables: (default " lvar ") ") nil nil lvar))
	)
    (setq lvar newlvar)
    (insert "ANY ")
    (insert lvar)
    (b-indent-line)
    (electric-b-terminate-line)
    (insert "WHERE")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (save-excursion
      (insert "preds?")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "THEN")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "subs?")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "END")
      (beginning-of-line)
      (b-indent-line)
      )
    )
  (end-of-line)
  )

(defun b-insert-let-block ()
  "Insert B LET ... BE ... IN ... END; block in the code with right indentation."
  (interactive)
  (save-excursion
    (let (
	  (newlvar (read-string (concat "Local variables: (default " lvar ") ") nil nil lvar))
	  )
      (setq lvar newlvar)
      (insert "LET ")
      (insert lvar)
      (insert " BE")
      (b-indent-line)
      (electric-b-terminate-line)
      (insert "preds?")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "IN")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "subs?")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "END")
      (beginning-of-line)
      (b-indent-line)
      )
    )
  (re-search-forward "\?")
  )

(defun b-insert-select-block ()
  "Insert B SELECT ... THEN ... WHEN ... THEN ... ELSE ... END; block in the code with right indentation."
  (interactive)
  (save-excursion
    (insert "SELECT preds? THEN")
    (b-indent-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "WHEN preds? THEN")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "ELSE")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "END")
    (beginning-of-line)
    (b-indent-line)
    )
  (re-search-forward "\?")
  )

(defun b-insert-choice-block ()
  "Insert B CHOICE ... OR ... END; block in the code with right indentation."
  (interactive)
  (save-excursion
    (insert "CHOICE")
    (b-indent-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "OR")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "END")
    (beginning-of-line)
    (b-indent-line)
    )
  (re-search-forward "\?")
  )

(defun b-insert-case-block ()
  "Insert B CASE ... OF ... EITHER ... THEN ... OR ... THEN ... ELSE ... END  ... END; block in the code with right indentation."
  (interactive)
  (save-excursion
    (insert "CASE frm? OF")
    (b-indent-line)
    (electric-b-terminate-line)
    (insert "EITHER frm? THEN")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "OR frm? THEN")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "ELSE")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "END")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "END")
    (beginning-of-line)
    (b-indent-line)
    )
  (re-search-forward "\?")
  )

(defun b-insert-var-block ()
  "Insert B VAR ... IN  ... END ; statement in the code with right indentation."
  (interactive)
  (save-excursion
    (let (
	  (newlvar (read-string (concat "Local variables: (default " lvar ") ") nil nil lvar))
	  )
      (setq lvar newlvar)
      (insert "VAR")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert lvar)
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "IN")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "subs?")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "END;")
      (beginning-of-line)
      (b-indent-line)
      )
    )
  (re-search-forward "\?")
  )

(defun b-insert-while-block ()
  "Insert B WHILE ... DO ... VARIANT ... INVARIANT ... END; block in the code with right indentation."
  (interactive)
  (insert "WHILE")
  (b-indent-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "preds?")
    (b-indent-line)
    (electric-b-terminate-line)
    (insert "DO")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "VARIANT")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "frm?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "INVARIANT")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "preds?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "END")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-elsif-substatement ()
  "Insert B ELSIF ... THEN ... ; sub-statement in the code with right indentation."
  (interactive)
  (save-excursion
    (insert "ELSIF")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "preds?")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "THEN")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    )
  (re-search-forward "\?")
  )

(defun b-insert-when-substatement ()
  "Insert B WHEN ... THEN ... ; sub-statement in the code with right indentation."
  (interactive)
  (save-excursion
    (insert "WHEN preds? THEN")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    )
  (re-search-forward "\?")
  )

(defun b-insert-or-substatement ()
  "Insert B OR ...  ; sub-statement in the code with right indentation."
  (interactive)
  (save-excursion
    (insert "OR")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    )
  (re-search-forward "\?")
  )

(defun b-insert-or-then-substatement ()
  "Insert B OR ... THEN ... ; sub-statement in the code with right indentation."
  (interactive)
  (save-excursion
    (insert "OR preds? THEN")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    )
  (re-search-forward "\?")
  )

(defun b-insert-op-block ()
  "Insert B template for an operation in the code with right indentation."
  (interactive)
  (save-excursion
    (let (
	  (opname (read-string (concat "Operation name: (default Op) ") nil nil "Op"))
	  (opinput (read-string (concat "Operation input(s): (default Nil) ") nil nil "?"))
	  (opoutput (read-string (concat "Operation output(s): (default Nil) ") nil nil "?"))
	  )
      (insert "/*\"")
      (end-of-line)
      (newline)
      (insert "\\textit{Operation:} \\textsf{")
      (insert (latex-string opname))
      (insert "}\\\\")
      (end-of-line)
      (newline)
      (insert "\\textit{Requirements:} \\\\")
      (end-of-line)
      (newline)
      (insert "\\textit{Precondition:}")
      (end-of-line)
      (newline)
      (insert "\"*/")
      (end-of-line)
      (newline)
      (if (not (string= opoutput "?"))
	  (insert (concat opoutput " <-- "))
	  )
      (insert opname)
      (if (not (string= opinput "?"))
	  (save-excursion
	     (insert "(")
	     (insert opinput)
	     (insert ")")
	     (insert " =")
	     )
	(insert " =")
	)
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "PRE")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "preds?")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "THEN")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "subs?")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (insert "END;")
      (beginning-of-line)
      (b-indent-line)
      (end-of-line)
      (electric-b-terminate-line)
      (forward-line -11)
      (b-indent-line)
      (forward-line 1)
      (b-indent-line)
      (forward-line 1)
      (b-indent-line)
      (forward-line 1)
      (b-indent-line)
      (forward-line 1)
      (b-indent-line)
      (forward-line 1)
      )
    )
    (forward-line -8)
    (beginning-of-line)
    (re-search-forward "\?")
  )

(defun latex-string (str) 
  "Convert a normal string to a latex string."
  (interactive)
  (save-excursion
    (setq len (length str))
    (setq index 0)
    (setq string "")
    (while (< index len)
      (setq ch (aref str index))
      (if (char= ch ?_)
	  (setq string (concat string "\\" (char-to-string ch)))
	(setq string (concat string (char-to-string ch)))
	)
      (setq index (+ index 1))
      )
    (setq newstring string)
    )
  )

(defun b-insert-set-block ()
  "Insert B template for declaring sets in the code with right indentation."
  (interactive)
  (insert "SETS")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "sets?")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-constant-block ()
  "Insert B template for declaring constants in the code with right indentation."
  (interactive)
  (insert "CONSTANTS")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "csts?")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-property-block ()
  "Insert B template for declaring properties in the code with right indentation."
  (interactive)
  (insert "PROPERTIES")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "preds?")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-variable-block ()
  "Insert B template for declaring variables in the code with right indentation."
  (interactive)
  (insert "VARIABLES")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "vrbs?")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-invariant-block ()
  "Insert B template for declaring variables in the code with right indentation."
  (interactive)
  (insert "INVARIANT")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "preds?")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-initialisation-block ()
  "Insert B template for declaring initialisation in the code with right indentation."
  (interactive)
  (insert "INITIALISATION")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "subs?")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-def-block ()
  "Insert B template for declaring definitions in the code with right indentation."
  (interactive)
  (insert "DEFINITIONS")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "frm? == frm?")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-see-block ()
  "Insert B template for declaring seen information in the code with right indentation."
  (interactive)
  (insert "SEES")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "mchs?")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-use-block ()
  "Insert B template for declaring used information in the code with right indentation."
  (interactive)
  (insert "USES")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "mchs?")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-include-block ()
  "Insert B template for declaring included information in the code with right indentation."
  (interactive)
  (insert "INCLUDES")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "mchs?(prms?)")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (electric-b-terminate-line)
    (insert "PROMOTES")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "opns?")
    (beginning-of-line)
    (b-indent-line)
    )
  (beginning-of-line)
  (re-search-forward "\?")
  )

(defun b-insert-extend-block ()
  "Insert B template for declaring extended information in the code with right indentation."
  (interactive)
  (insert "EXTENDS")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "mchs?(prms?)")
    (beginning-of-line)
    (b-indent-line)
    )
  (beginning-of-line)
  (re-search-forward "\?")
  )

(defun b-insert-machine-block ()
  "Insert B template for declaring machine in the code with right indentation."
  (interactive)
  (insert "MACHINE Name?(prms?)")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "CONSTRAINTS")
    (beginning-of-line)
    (b-indent-line)
    (end-of-line)
    (electric-b-terminate-line)
    (insert "preds?")
    (beginning-of-line)
    (b-indent-line)
    )
  (beginning-of-line)
  (re-search-forward "\?")
  )

(defun b-insert-import-block ()
  "Insert B template for declaring machine in the code with right indentation."
  (interactive)
  (insert "IMPORTS")
  (beginning-of-line)
  (b-indent-line)
  (end-of-line)
  (electric-b-terminate-line)
  (save-excursion
    (insert "mchs?")
    (beginning-of-line)
    (b-indent-line)
    )
  (end-of-line)
  )

(defun b-insert-unprint-comment-block ()
  "Insert B template for unprintable comment block with right indentation."
  (interactive)
  (save-excursion
    (forward-line -1)
    (end-of-line)
    (newline)
    (insert "/*")
    (end-of-line)
    (newline)	
    (insert "unprintable comments?")
    (end-of-line)
    (newline)
    (insert "*/")
    (end-of-line)
    (forward-line -3)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    )
  (forward-line -2)
  (end-of-line)
  )

(defun b-insert-print-comment-block ()
  "Insert B template for unprintable comment block with right indentation."
  (interactive)
  (save-excursion
    (forward-line -1)
    (end-of-line)
    (newline)
    (insert "/*\"")
    (end-of-line)
    (newline)
    (insert "printable comments?")
    (end-of-line)
    (newline)
    (insert "\"*/")
    (end-of-line)
    (forward-line -3)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    )
  (forward-line -2)
  (end-of-line)
  )

(defun b-insert-operation-comment-block ()
  "Insert B template for unprintable comment block with right indentation."
  (interactive)
  (save-excursion
    (forward-line -1)
    (end-of-line)
    (newline)
    (insert "/*\"")
    (end-of-line)
    (newline)
    (insert "\\textit{Operation:} \\textsf{Op}\\\\")
    (end-of-line)
    (newline)
    (insert "\\textit{Requirements:} \\\\")
    (end-of-line)
    (newline)
    (insert "\\textit{Precondition:}")
    (end-of-line)
    (newline)
    (insert "\"*/")
    (end-of-line)
    (forward-line -5)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    )
  (forward-line -4)
  (end-of-line)
  )

(defun b-insert-header-comment-block ()
  "Insert B template for unprintable comment block with right indentation."
  (interactive)
  (save-excursion
    (goto-char 1)
    (insert "/*")
    (newline)
    (setq name (buffer-name))
    (insert name)
    (newline)
    (insert "Author: ")
    (insert author)
    (newline)
    (insert "Email: ")
    (insert email)
    (newline)
    (setq time (current-time-string))
    (insert "Create: ")
    (insert time)
    (newline)
    (insert "Last modified: ")
    (insert time)
    (newline)
    (newline)
    (insert "*/")
    (newline)
    (newline)
    (forward-line -8)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    (forward-line 1)
    (b-indent-line)
    )
  (goto-line 7)
  (end-of-line)
  )

(defun b-save-buffer ()
  "Set the modified time and save the buffer."
  (interactive)
  (save-excursion
    (setq time (current-time-string))
    (goto-char 0)
    (if (buffer-modified-p)
	(if (re-search-forward "Last modified: " nil t)
	    (save-excursion
	      (setq beg (point))
	      (end-of-line)
	      (setq end (point))
	      (delete-region beg end)
	      (insert time)
	      (save-buffer)
	      )
	  (save-excursion
	    (b-insert-header-comment-block)
	    )
	  )
      (message "(No changes need to be saved)")
      )
    )
  )

;;;;
;;;; Indentation
;;;;

(defun b-indent-line ()
  "Indent current line as a B statement."
  (save-excursion
    (let* ((indent-str (b-calculate-indent))
	   (type (car indent-str))
	   (ind (car (cdr indent-str))))
      (if (eq type 'construct)
	  (save-excursion 
	    (beginning-of-line)
	    (delete-horizontal-space)
	    )
	(if (eq type 'op)
	    (save-excursion
	      (beginning-of-line)
	      (delete-horizontal-space)
	      (indent-to b-indent-level)
	      )	
	  (if (eq type 'same)
	      (save-excursion
		(beginning-of-line)
		(delete-horizontal-space)
		(indent-to ind)
		)
	    (if (eq type 'out)
		(save-excursion
		  (beginning-of-line)
		  (delete-horizontal-space)
		  (indent-to (- ind b-indent-level))
		  )
	      (if (eq type 'in)
		  (save-excursion
		    (beginning-of-line)
		    (delete-horizontal-space)
		    (indent-to (+ ind b-indent-level))
		    )
		(if (eq type 'comment)
		    (save-excursion
		      (message "Comment")
		      (b-indent-comment)
		      )
		  (if (eq type 'end)
		      (save-excursion
			(message "End comment")
			(b-indent-end-comment)
			)
		    (if (eq type 'begin)
			(save-excursion
;			  (message "Begin comment")
			  (b-indent-begin-comment)
			  )
		      )
		    )
		  )
		)
	      )
	    )
	  )
	)
      )
    )
  )

(defun b-calculate-indent ()
  "Calculate the indent of the current B line.
   ;;Return a list of two elements: (INDENT-TYPE INDENT-LEVEL)."
  (save-excursion
    (let* ((comment (car (b-inside-comment)))
	   (type (catch 'nesting
		   (cond

		    (
		     (looking-at b-invariant-re)
		     (throw 'nesting 'out)
		     )

		    (
		     (looking-at b-begin-comment-re)
		     (throw 'nesting 'begin)
		     )

		    (
		     (looking-at b-end-comment-re)
		     (throw 'nesting 'end)
		     )

		    (
		     (eq comment 'yes)
		     (throw 'nesting 'comment)
		     )

		    (
		     (looking-at b-construct-re) 
		     (throw 'nesting 'construct)
		     )
		    
		    (
		     (or (looking-at b-op-param-out-re)
			 (looking-at b-op-out-re)
			 (looking-at b-op-param-re)
			 (looking-at b-op-re)
			 )
		     (throw 'nesting 'op)
		     )
		    
		    (
		     (or (looking-at b-then-re) 
			 (looking-at b-end-block-re)
			 (looking-at b-reason-re)
;			 (looking-at b-proof-rule-re)
			 )
		     (throw 'nesting 'out)
		     )

		    (
		     (or (looking-at b-where-re)
			 (looking-at "^[ \t]*;[ \t]*$")
			 (looking-at "^[ \t]*,[ \t]*$")
			 )
		     (throw 'nesting 'same)
		     )

		    ;; look at the last line
		    (
		     (if (save-excursion
			   (forward-line -1)
			   (beginning-of-line)
			   (while (looking-at b-empty-line-re)
			     (forward-line -1)
			     )
			   (while (looking-at b-end-comment-re)
			     (re-search-backward b-begin-comment-re nil t)
			     (forward-line -1)
			     (while (looking-at b-empty-line-re)
			       (forward-line -1)
			       )
			     )
			   (and
			    (not (or (looking-at b-construct-re)
				     (looking-at b-begin-block-re)
				     (looking-at b-then-re)
				     (looking-at b-pchoice-re)
				     (looking-at b-begin-comment-re)
				     ))
			    (or (looking-at b-op-re)
				(looking-at b-op-param-out-re)
				(looking-at b-op-param-re)
				(looking-at b-op-out-re)
				(looking-at b-end-block-re)
				(looking-at "^.*?\\(or\\|;\\|&\\|,\\|||\\)[ \t]*$")
				)
			    )
			   )
			 (throw 'nesting 'same)
		       (throw 'nesting 'in)
		       )
		     )
		    )
		   )
		 )
	   )
      (list type (b-last-indent))
      )
    )
  )


(defun b-last-indent ()
  "Return the indent-level the current statement has.
   Do not count labels, case-statements or records."
  (save-excursion
    (forward-line -1)
    (while (looking-at b-empty-line-re)
      (forward-line -1)
      )
    (while (looking-at b-end-comment-re)
      (re-search-backward b-begin-comment-re nil t)
      (forward-line -1)
      (while (looking-at b-empty-line-re)
	(forward-line -1)
	)
      )
    (beginning-of-line)
    (skip-chars-forward " \t")
    (current-column)
    )
  )

(defun b-indent-comment (&optional arg)
  "Indent current line as comment.
   If optional arg is non-nil, just return the
   column number the line should be indented to."
  (let* ((stcol (save-excursion
		  (re-search-backward b-begin-comment-re nil t)
		  (beginning-of-line)
		  (skip-chars-forward " \t")
		  (+ b-indent-level (current-column))
		  )
		)
	 )
    (if arg stcol
      (beginning-of-line)
      (delete-horizontal-space)
      (indent-to stcol)
      )
    )
  )

(defun b-indent-end-comment (&optional arg)
  "Indent current line as comment.
   If optional arg is non-nil, just return the
   column number the line should be indented to."
  (let* ((stcol (save-excursion
		  (re-search-backward b-begin-comment-re nil t)
		  (beginning-of-line)
		  (skip-chars-forward " \t")
		  (current-column)
		  )
		)
	 )
    (if arg stcol
      (beginning-of-line)
      (delete-horizontal-space)
      (indent-to stcol)
      )
    )
  )

(defun b-indent-begin-comment (&optional arg)
  "Indent current line as comment.
   If optional arg is non-nil, just return the
   column number the line should be indented to."
  (let* ((oldpos (point)))
    (save-excursion
      (while (looking-at b-begin-comment-re)
	(re-search-forward b-end-comment-re nil t)
	(forward-line 1)
	(while (looking-at b-empty-line-re)
	  (forward-line 1)
	  )
	(setq indent-str (b-calculate-indent))
	(setq type (car indent-str))
	(setq ind (car (cdr indent-str)))
	(if (eq type 'construct)
	    (save-excursion 
	      (goto-char oldpos)
	      (beginning-of-line)
	      (delete-horizontal-space)
	      )
	  (if (eq type 'op)
	      (save-excursion
		(goto-char oldpos)
		(beginning-of-line)
		(delete-horizontal-space)
		(indent-to b-indent-level)
		)	
	    (if (eq type 'same)
		(save-excursion
		  (goto-char oldpos)
		  (beginning-of-line)
		  (delete-horizontal-space)
		  (indent-to ind)
		  )
	      (if (eq type 'out)
		  (save-excursion
		    (goto-char oldpos)
		    (beginning-of-line)
		    (delete-horizontal-space)
		    (indent-to (- ind b-indent-level))
		    )
		(if (eq type 'in)
		    (save-excursion
		      (goto-char oldpos)
		      (beginning-of-line)
		      (delete-horizontal-space)
		      (indent-to (+ ind b-indent-level))
		      (message "Indent to %d" ind)
		      )
		  )
		)
	      )
	    )
	  )
	)
      )
    )
  )

(defun b-inside-comment ()
  "Check if the current position is inside a comment or not"
  (save-excursion
    (let* ((oldpos (point))
	   (begin (if
		      (re-search-backward b-begin-comment-re nil t)
		      (point)
		    ()
		    )
		  )
	   (end (if
		    (re-search-forward b-end-comment-re nil t)
		    (point)
		  ()
		  )
		)
	   (answer (catch 'comment
		     (if 
			 (and begin
			      end
			      (> oldpos begin)
			      (< oldpos end)
			      )
			 (throw 'comment 'yes)
		       (throw 'comment 'no)
		       )
		     )
		   )
	   )
      (list answer)
      )
    )
  )

(defun b-inside-print-comment ()
  "Check if the current position is inside a comment or not"
  (save-excursion
    (let* ((oldpos (point))
	   (begin (if
		      (re-search-backward b-begin-print-comment-re nil t)
		      (point)
		    ()
		    )
		  )
	   (end (if
		    (re-search-forward b-end-print-comment-re nil t)
		    (point)
		  ()
		  )
		)
	   (answer (catch 'comment
		     (if 
			 (and begin
			      end
			      (> oldpos begin)
			      (< oldpos end)
			      )
			 (throw 'comment 'yes)
		       (throw 'comment 'no)
		       )
		     )
		   )
	   )
      (list answer)
      )
    )
  )

(defun b-latex-underscore ()
  "Insert \\_ if inside printable comment, otherwise insert normal _."
  (interactive)
  (let* ((comment (car (b-inside-print-comment))))
    (if (eq comment 'yes)
	(insert "\\_")
      (insert "_")
      )
    )
  )

(defun b-latex-doublequote ()
  "Insert '' if inside printable comment, otherwise insert normal \"."
  (interactive)
  (let* ((comment (car (b-inside-print-comment))))
    (if (eq comment 'yes)
	(insert "''")
      (insert "\"")
      )
    )
  )

(defun b-latex-backslash ()
  "Insert latex macro if inside printable comment, otherwise insert normal \\."
  (interactive)
  (let* ((comment (car (b-inside-print-comment))))
    (if (eq comment 'yes)
	(save-excursion
	  (let (
		(newtexmacro (read-string (concat "Macro (default " texmacro "): \\") nil nil texmacro))
		)
	    (setq texmacro newtexmacro)
	    (insert "\\")
	    (insert texmacro)
	    (if (not (string= texmacro "\\"))
		(insert "{}")
	      )
	    )
	  )
      (insert "\\")
      )
    (if (eq comment 'yes)
	(if (not (string= texmacro "\\"))
	    (re-search-forward "{")
	  (forward-char 2)
	  )
      )
    )
  )

(defun b-latex-space ()
  "Auto wrapping the line if inside printable comment, otherwise insert normal space."
  (interactive)
  (insert " ")
  (save-excursion
    (re-search-backward " ")
    (let* ((comment (car (b-inside-comment))))
      (if (eq comment 'yes)
	  (let (
		(width (frame-width))
;		(column (current-column))
		)
	    (setq width (- width 2))
	    (setq column (current-column))
	    (if (> column width)
		(save-excursion
		  (setq over 'yes)
		  (while (> column width)
		    (re-search-backward " ")
		    (setq column (current-column))
		  )
		  (newline)
		  (forward-line -1)
		  (b-indent-line)
		  (forward-line 1)
		  (b-indent-line)
		  )
	      (setq over 'no)
	      )
	    )
	)
      )
    )
  )

;; b.el ends here

